"""
Quantiphyse - Widgets for data smoothing

Copyright (c) 2013-2018 University of Oxford
"""

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, TitleWidget, OverlayCombo, NumericOption
from quantiphyse.utils import QpException

from .process import SmoothingProcess
    
class SmoothingWidget(QpWidget):
    """
    Gaussian smoothing widget
    """
    def __init__(self, **kwargs):
        super(SmoothingWidget, self).__init__(name="Smoothing", icon="smooth.png", 
                                              desc="Gaussian smoothing", group="Processing", **kwargs)
        
    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="Data Smoothing", subtitle="Smooth data using a Gaussian kernel", help="smoothing")
        vbox.addWidget(title)

        hbox = QtGui.QHBoxLayout()
        gbox = QtGui.QGroupBox()
        gbox.setTitle("Options")
        grid = QtGui.QGridLayout()
        gbox.setLayout(grid)

        grid.addWidget(QtGui.QLabel("Data to smooth"), 0, 0)
        self.data_combo = OverlayCombo(self.ivm)
        self.data_combo.currentIndexChanged.connect(self.data_changed)
        grid.addWidget(self.data_combo, 0, 1)
        self.sigma = NumericOption("Sigma (mm)", grid, xpos=0, ypos=1, minval=0, step=0.1, default=1.0)
        grid.addWidget(QtGui.QLabel("Output name"), 2, 0)
        self.output_name = QtGui.QLineEdit()
        grid.addWidget(self.output_name, 2, 1)

        hbox.addWidget(gbox)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        hbox = QtGui.QHBoxLayout()
        self.run_btn = QtGui.QPushButton("Run")
        self.run_btn.clicked.connect(self.run)
        hbox.addWidget(self.run_btn)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addStretch(1) 
        self.data_changed()  
    
    def data_changed(self):
        self.output_name.setText("%s_smoothed" % self.data_combo.currentText())
        self.run_btn.setEnabled(self.data_combo.currentText() in self.ivm.data)

    def batch_options(self):
        return "Smooth", {"data" : self.data_combo.currentText(),
                          "sigma" : self.sigma.spin.value(),
                          "output-name" : self.output_name.text()}

    def run(self):
        process = SmoothingProcess(self.ivm)
        process.run(self.batch_options()[1])



"""
Quantiphyse - Histogram widget

Copyright (c) 2013-2018 University of Oxford
"""
try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, TitleWidget
from quantiphyse.gui.plot import Plot
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, BoolOption, ChoiceOption

from .process import HistogramProcess
    
class HistogramWidget(QpWidget):
    """
    Widget which displays data histograms
    """
    def __init__(self, **kwargs):
        super(HistogramWidget, self).__init__(name="Histogram", icon="hist",
                                              desc="Display histograms from data", group="Visualisation", **kwargs)
        self._updating = False

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self)
        vbox.addWidget(title)

        self.options = OptionBox("Options")
        self.options.add("Data", DataOption(self.ivm, multi=True), key="data")
        self.options.add("Within ROI", DataOption(self.ivm, data=False, rois=True, none_option=True), key="roi")
        self.options.add("All volumes", BoolOption(default=False), key="allvols")
        self.options.add("Y-axis scale", ChoiceOption(["Count", "Probability"]), key="yscale")
        self.options.add("Number of bins", NumericOption(minval=5, maxval=500, default=100, intonly=True), key="bins")
        self.options.add("Min value", NumericOption(minval=0, maxval=100, default=0), key="min")
        self.options.add("Max value", NumericOption(minval=0, maxval=500, default=100), key="max")
        self.options.option("yscale").sig_changed.connect(self._yscale_changed)
        self.options.option("min").sig_changed.connect(self._min_changed)
        self.options.option("min").sig_changed.connect(self._max_changed)
        vbox.addWidget(self.options)

        self.plot = Plot(qpo=None, parent=self, title="Data histogram", display_mode=False)
        self.plot.set_xlabel("Data value")
        self.plot.set_ylabel("Count")
        vbox.addWidget(self.plot)

        vbox.addStretch(1)

    def activate(self):
        self.ivm.sig_all_data.connect(self._data_changed)
        self.options.option("data").sig_changed.connect(self._data_changed)
        self.options.sig_changed.connect(self._update)
        self._data_changed()

    def deactivate(self):
        self.ivm.sig_all_data.disconnect(self._data_changed)
        self.options.option("data").sig_changed.disconnect(self._data_changed)
        self.options.sig_changed.disconnect(self._update)

    def processes(self):
        opts = self.options.values()
        if not opts.pop("allvols", False):
            opts["vol"] = self.ivl.focus()[3]

        return {
            "Histogram" : opts
        }

    def _yscale_changed(self):
        self.plot.set_ylabel(self.options.option("yscale").value)

    def _min_changed(self):
        minval = self.options.option("min").value
        self.options.option("max").setLimits(minval=minval)

    def _max_changed(self):
        maxval = self.options.option("max").value
        self.options.option("min").setLimits(maxval=maxval)

    def _data_changed(self):
        if self._updating: return
        self._updating = True
        try:
            data_names = self.options.option("data").value
            vol = None
            if self.options.option("allvols").value:
                vol = self.ivl.focus()[3]
            dmin, dmax, multivol = None, None, False
            for data_name in data_names:
                qpdata = self.ivm.data[data_name]
                multivol = multivol or qpdata.nvols > 1
                _dmin, _dmax = qpdata.range(vol=vol)
                if dmin is None or dmin > _dmin:
                    dmin = _dmin
                if dmax is None or dmax > dmax:
                    dmax = _dmax
            
            if dmin is not None and dmax is not None:
                self.options.option("min").value = dmin
                self.options.option("min").setLimits(dmin, dmax)
                self.options.option("max").value = dmax
                self.options.option("max").setLimits(dmin, dmax)
                
            self.options.set_visible("allvols", multivol)
            self._update()
        finally:
            self._updating = False

    def _update(self):
        opts = self.processes()["Histogram"]
        if opts["data"]:
            HistogramProcess(self.ivm).run(opts)
            self.plot.clear()
            histogram = self.ivm.extras.get("histogram", None)
            if histogram is not None:
                for idx, name in enumerate(histogram.col_headers[3:]):
                    xvalues = [row[2] for row in histogram.arr]
                    yvalues = [row[idx+3] for row in histogram.arr]
                    self.plot.add_line(yvalues, name=name, xvalues=xvalues)



"""
Quantiphyse - Base class for process self-test framework

Copyright (c) 2013-2018 University of Oxford
"""

import sys
import os
import math
import unittest
import traceback
import tempfile
import shutil
import time

import numpy as np
import scipy
import nibabel as nib

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import DataGrid, ImageVolumeManagement
from quantiphyse.processes import Process
from quantiphyse.utils.batch import Script
from quantiphyse.utils import QpException

class ProcessTest(unittest.TestCase):
    """
    Base class for a test module for a QP ProcessTest

    Process tests consist of running example batch scripts
    and checking the output. A set of test data files are created 
    in a temporary directory ``self.test_data_path``
    """
    def setUp(self):
        self.qpe, self.error = False, False
        sys.excepthook = self._exc

        self.ivm = ImageVolumeManagement()
        self.input_dir = tempfile.mkdtemp(prefix="qp")
        self.output_dir = tempfile.mkdtemp(prefix="qp")
        self._create_test_data_files()
        self.status, self.log, self.exception = None, None, None

    def tearDown(self):
        shutil.rmtree(self.input_dir)
        shutil.rmtree(self.output_dir)
            
    def run_yaml(self, yaml):
        script = Script(self.ivm)
        script.sig_finished.connect(self._script_finished)

        full_yaml = """
OutputFolder: %s
InputFolder: %s

Processing:
  - Load:
        data:
            data_3d.nii.gz:
            data_4d.nii.gz:
            data_4d_moving.nii.gz:
        rois:
            mask.nii.gz:
"""  % (self.output_dir, self.input_dir) + yaml + """
Cases:
  - case:
"""
        script.execute({"yaml" : full_yaml})
        while script.status == Script.RUNNING:
            self.processEvents()
            time.sleep(1)
        if self.status != Script.SUCCEEDED:
            raise self.exception 

    def _script_finished(self, *args):
        self.status, self.log, self.exception = args

    def processEvents(self):
        """
        Process outstanding QT events, i.e. let handlers for widget
        events that we have triggered run

        This must be run every time a test triggers widget events
        in order for the test to detect the effects
        """
        QtCore.QCoreApplication.instance().processEvents()

    def _create_test_data_files(self):
        """
        Create test data files

        This is done freshly for each test in case the data is modified
        during the process of a test
        """
        from . import create_test_data
        testshape = getattr(self, "testshape", (10, 10, 10))
        create_test_data(self, shape=testshape)

        nii_3d = nib.Nifti1Image(self.data_3d, np.identity(4))
        nii_3d.to_filename(os.path.join(self.input_dir, "data_3d.nii.gz"))

        nii_4d = nib.Nifti1Image(self.data_4d, np.identity(4))
        nii_4d.to_filename(os.path.join(self.input_dir, "data_4d.nii.gz"))

        nii_4d_moving = nib.Nifti1Image(self.data_4d_moving, np.identity(4))
        nii_4d_moving.to_filename(os.path.join(self.input_dir, "data_4d_moving.nii.gz"))

        nii_mask = nib.Nifti1Image(self.mask, np.identity(4))
        nii_mask.to_filename(os.path.join(self.input_dir, "mask.nii.gz"))

    def _exc(self, exc_type, value, tb):
        """ 
        Exception handler which simply flags whether a user-exception or an error has been caught 
        """
        self.qpe = issubclass(exc_type, QpException)
        self.error = not self.qpe
        if self.error or "--debug" in sys.argv:
            traceback.print_exception(exc_type, value, tb)
        



"""
Quantiphyse - Subclasses of QpData for handling Dicom data

Copyright (c) 2013-2018 University of Oxford
"""
from __future__ import division, print_function

import sys
import os
import warnings
import glob
import logging

import numpy as np

import nibabel as nib

HAVE_DCMSTACK = True
try:
    import dcmstack
    import dicom
except ImportError:
    HAVE_DCMSTACK = False
    warnings.warn("DCMSTACK not found - may not be able to read DICOM folders")

from quantiphyse.utils import QpException
from .qpdata import DataGrid, QpData, NumpyData

LOG = logging.getLogger(__name__)

class DicomFolder(QpData):
    """
    QpData instance loaded from a directory of DICOM files
    """
    def __init__(self, fname):
        # A directory containing DICOMs. Convert them to Nifti
        print("Converting DICOMS in %s..." % (os.path.basename(fname)))
        src_dcms = glob.glob('%s/*' % fname)
        nii = None
        try:
            if HAVE_DCMSTACK:
                # Give DCMSTACK a chance to do its thing
                stacks = dcmstack.parse_and_stack(src_dcms)
                stack = stacks.values()[0]
                nii = stack.to_nifti()
        except:
            warnings.warn("DCMSTACK failed - trying our method")
        if nii is None:
            # Try our top-secret in-house method
            nii = self.fallback_dcmstack(src_dcms)

        print("DONE\n")
        if len(nii.shape) > 3:
            nvols = nii.shape[3]
        else:
            nvols = 1

        grid = DataGrid(nii.shape[:3], nii.header.get_best_affine())
        self.dcmdata = nii.get_data()
        QpData.__init__(self, fname, grid, nvols, fname=fname)

    def raw(self):
        return self.dcmdata

    def fallback_dcmstack(self, fnames):
        """
        Create NIFTI from DCM files. Quick and dirty method
        for when DCMSTACK does not exist or fails.

        Basically we determine the sequence using the InstanceNumber tag
        but make sure we put slices together into volumes using the SliceLocation tag
        """
        import nibabel.nicom.dicomwrappers as nib_dcm
        ignored_files = []
        first = True
        dcms = []
        slices = set()
        ss, rs, ri = 1, 1, 0 # Pixel value scaling
        sys.stdout.write("  0%")
        for idx, fname in enumerate(fnames):
            try:
                dcm = dicom.read_file(fname)
            except:
                ignored_files.append(fname)
                continue
            slices.add(float(dcm.SliceLocation))
            dcms.append(dcm)

            if first:
                dcm1 = nib_dcm.wrapper_from_file(fname)
                dcm_affine = dcm1.get_affine()
                print(dcm_affine)
                try:
                    # Need all three of these to be of use
                    ss = dcm[0x2005, 0x100e].value
                    rs = dcm[0x2005, 0x140a].value
                    ri = dcm[0x2005, 0x1409].value
                except:
                    pass
                first = False

            percent = 100*float(idx+1) / len(fnames)
            sys.stdout.write("\b\b\b\b%3i%%" % int(percent))
            sys.stdout.flush()
        
        if len(slices) < 0:
            raise QpException("This doesn't seem to be a DICOM folder")
            
        n_vols = int(len(dcms) / len(slices))
        if n_vols * len(slices) != len(dcms):
            raise QpException("Could not parse DICOMS - unable to determine fixed number of volumes")

        print("\n")
        print("%i Volumes" % n_vols)
        print("Ignored (non-DICOM) files: %i" % len(ignored_files))
        print("Slice locations are: " + ", ".join([str(s) for s in slices]))
        print("RescaleSlope: %f" % rs)
        print("RescaleIntercept: %f" % ri)
        print("ScaleSlope: %f" % ss)
        for sidx, s in enumerate(sorted(slices)):
            for idx, dcm in enumerate(sorted([d for d in dcms if d.SliceLocation == s], key=lambda x: x.InstanceNumber)):
                dcm.GlobalIndex = sidx + len(slices) * idx
            
        print("Creating NIFTI...")
        data = np.zeros([dcm1.image_shape[0], dcm1.image_shape[1], len(slices), int(n_vols)])
        sidx, vidx = 0, 0
        for dcm in sorted(dcms, key=lambda x: x.GlobalIndex):
            data[:, :, sidx, vidx] = (np.squeeze(dcm.pixel_array)*rs + ri)/ss
            sidx += 1
            if sidx == len(slices):
                sidx = 0
                vidx += 1

        nii = nib.Nifti1Image(data, dcm_affine)
        nii.update_header()
        print("DONE")
        return nii



import unittest

import numpy as np

from quantiphyse.test.widget_test import WidgetTest

from .widget import PcaWidget

NUM_PCA = 5
NAME = "pca"

class PcaWidgetTest(WidgetTest):

    def widget_class(self):
        return PcaWidget

    def testNoData(self):
        if self.w._run.runBtn.isEnabled():
            self.w._run.runBtn.clicked.emit()
        self.assertFalse(self.error)

    def test3dData(self):
        """ 4D data required """
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d") 
        self.assertFalse(bool(self.w._options.option("data").value))
        
    def test4dData(self):
        """ 4d clustering """
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        self.processEvents()            
        self.w._options.option("data").value = "data_4d"
        self.processEvents()            
        self.assertTrue(self.w._run.runBtn.isEnabled())
        self.assertEquals(self.w._options.option("output-name").value, "data_4d_pca")
        
        self.w._options.option("n-components").value = NUM_PCA
        self.w._options.option("output-name").value = NAME
        self.w._run.runBtn.clicked.emit()
        self.processEvents()
        
        for mode in range(NUM_PCA):
            self.assertTrue("%s%i" % (NAME, mode) in self.ivm.data)
        self.assertFalse("%s%i" % (NAME, NUM_PCA+1) in self.ivm.data)

        self.assertEquals(self.ivm.current_data.name, "%s0" % NAME)
        self.assertFalse(self.error)
        
    def test4dDataWithRoi(self):
        """ 4d clustering within an ROI"""
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        self.ivm.add(self.mask, grid=self.grid, name="mask")
        self.w._options.option("data").value = "data_4d"
        self.w._options.option("roi").value = "mask"
        
        self.assertTrue(self.w._run.runBtn.isEnabled())
        self.assertEquals(self.w._options.option("output-name").value, "data_4d_pca")
        
        self.w._options.option("n-components").value = NUM_PCA
        self.w._options.option("output-name").value = NAME
        self.w._run.runBtn.clicked.emit()
        self.processEvents()
        
        for mode in range(NUM_PCA):
            name = "%s%i" % (NAME, mode)
            self.assertTrue(name in self.ivm.data)
            # Data value is always zero outside the ROI
            data = self.ivm.data[name].raw()
            self.assertTrue(np.all(data[self.mask == 0] == 0))
            self.assertFalse(self.error)

        self.assertFalse("%s%i" % (NAME, NUM_PCA+1) in self.ivm.data)
        self.assertEquals(self.ivm.current_data.name, "%s0" % NAME)
        self.assertFalse(self.error)
        
if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Processes for basic loading/saving of data

Copyright (c) 2013-2018 University of Oxford
"""

import os

from quantiphyse.utils import QpException
from quantiphyse.data import load, save

from .process import Process

__all__ = ["LoadProcess", "LoadDataProcess", "LoadRoisProcess", "SaveProcess", "SaveAllExceptProcess", "SaveDeleteProcess", "SaveArtifactsProcess"]

class LoadProcess(Process):
    """
    Load data into the IVM
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        rois = options.pop('rois', {})
        data = options.pop('data', {})
        # Force 3D data to be multiple 2D volumes 
        force_mv = options.pop('force-multivol', False)

        for fname, name in list(data.items()) + list(rois.items()):
            qpdata = self._load_file(fname, name)
            if qpdata is not None: 
                if force_mv and qpdata.nvols == 1 and qpdata.grid.shape[2] > 1: 
                    qpdata.set_2dt()
                qpdata.roi = fname in rois
                self.ivm.add(qpdata, make_current=True)

    def _load_file(self, fname, name):
        filepath = self._get_filepath(fname)
        if name is None:
            name = self.ivm.suggest_name(os.path.split(fname)[1].split(".", 1)[0])
        self.debug("  - Loading data '%s' from %s" % (name, filepath))
        try:
            data = load(filepath)
            data.name = name
            return data
        except QpException as exc:
            self.warn("Failed to load data: %s (%s)" % (filepath, str(exc)))

    def _get_filepath(self, fname, folder=None):
        if os.path.isabs(fname):
            return fname
        else:
            if folder is None: folder = self.indir
            return os.path.abspath(os.path.join(folder, fname))

class LoadDataProcess(LoadProcess):
    """
    Process to load data

    Deprecated: use LoadProcess
    """
    def run(self, options):
        LoadProcess.run(self, {'data' : options})
        for key in list(options.keys()): options.pop(key)

class LoadRoisProcess(LoadProcess):
    """
    Process to load ROIs

    Deprecated: use LoadProcess
    """
    def run(self, options):
        LoadProcess.run(self, {'rois' : options})
        for key in list(options.keys()): options.pop(key)

class SaveProcess(Process):
    """
    Save data to file
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        # Note that output-grid is not a valid data name so will not clash
        output_grid = None
        output_grid_name = options.pop("output-grid", None)
        if output_grid_name is not None:
            output_grid_data = self.ivm.data.get(output_grid_name, self.ivm.rois.get(output_grid_name, None))
            if output_grid_data is None:
                raise QpException("No such data found as source of grid: %s" % output_grid_name)
            else:
                output_grid = output_grid_data.grid

        for name in list(options.keys()):
            try:
                fname = options.pop(name, name)
                qpdata = self.ivm.data.get(name, None)
                if qpdata is not None:
                    save(qpdata, fname, grid=output_grid, outdir=self.outdir)
                else:
                    self.warn("Failed to save %s - no such data or ROI found" % name)
            except QpException as exc:
                self.warn("Failed to save %s: %s" % (name, str(exc)))

class SaveAllExceptProcess(Process):
    """
    Save all data to file apart from specified items
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        exceptions = list(options.keys())
        for k in exceptions: options.pop(k)

        for name, qpdata in self.ivm.data.items():
            if name in exceptions: 
                continue
            try:
                save(qpdata, name, outdir=self.outdir)
            except QpException as exc:
                self.warn("Failed to save %s: %s" % (name, str(exc)))
            except:
                import traceback
                traceback.print_exc()

class SaveDeleteProcess(SaveProcess):
    """
    Save data to file and then delete it
    """
    def __init__(self, ivm, **kwargs):
        SaveProcess.__init__(self, ivm, **kwargs)

    def run(self, options):
        options_save = dict(options)
        SaveProcess.run(self, options)

        for name in options_save:
            if name in self.ivm.data: self.ivm.delete(name)

class SaveArtifactsProcess(Process):
    """
    Save 'extras' (previously known as 'artifacts')
    """
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        for name in list(options.keys()):
            fname = options.pop(name)
            if not fname: fname = name
            if name in self.ivm.extras: 
                self.debug("Saving '%s' to %s" % (name, fname))
                self._save_text(str(self.ivm.extras[name]), fname)
            else:
                self.warn("Extra '%s' not found - not saving" % name)

    def _save_text(self, text, fname, ext="txt"):
        if text:
            if "." not in fname: fname = "%s.%s" % (fname, ext)
            if not os.path.isabs(fname):
                fname = os.path.join(self.outdir, fname)
            dirname = os.path.dirname(fname)
            if not os.path.exists(dirname): os.makedirs(dirname)
            with open(fname, "w") as text_file:
                text_file.write(text)



"""
Quantiphyse - Basic data processes

Copyright (c) 2013-2018 University of Oxford
"""

from .process import Process
from .feat_pca import PcaFeatReduce as PCA
from . import normalisation

__all__ = ["Process", "PCA", "normalisation"]



"""
Quantiphyse - Analysis processes for data simulation

Copyright (c) 2013-2018 University of Oxford
"""
import math

import numpy as np
import scipy.ndimage.interpolation

from quantiphyse.data import DataGrid
from quantiphyse.utils import QpException
from quantiphyse.processes import Process

class AddNoiseProcess(Process):
    """
    Simple process for adding gaussian noise
    """
    PROCESS_NAME = "AddNoise"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data = self.get_data(options)

        output_name = options.pop("output-name", "%s_noisy" % data.name)
        std = float(options.pop("std"))

        noise = np.random.normal(loc=0, scale=std, size=list(data.grid.shape) + [data.nvols,])
        if data.nvols == 1: 
            noise = np.squeeze(noise, -1)
        noisy_data = data.raw() + noise
        self.ivm.add(noisy_data, grid=data.grid, name=output_name, make_current=True)

class SimMotionProcess(Process):
    """
    Simple process for adding gaussian noise
    """
    PROCESS_NAME = "SimMotion"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data = self.get_data(options)
        if data.ndim != 4:
            raise QpException("Can only simulate motion on 4D data")

        output_name = options.pop("output-name", "%s_moving" % data.name)
        std = float(options.pop("std"))
        std_voxels = [std / size for size in data.grid.spacing]
        output_grid = data.grid
        output_shape = data.grid.shape
            
        padding = options.pop("padding", 0)
        if padding > 0:
            padding_voxels = [int(math.ceil(padding / size)) for size in data.grid.spacing]
            # Need to adjust the origin so the output data lines up with the input
            output_origin = np.copy(data.grid.origin)
            output_shape = np.copy(data.grid.shape)
            output_affine = np.copy(data.grid.affine)
            for axis in range(3):
                output_origin[axis] -= np.dot(padding_voxels, data.grid.transform[axis, :])
                output_shape[axis] += 2*padding_voxels[axis]
            output_affine[:3, 3] = output_origin
            output_grid = DataGrid(output_shape, output_affine)

        moving_data = np.zeros(list(output_shape) + [data.nvols,])
        for vol in range(data.nvols):
            voldata = data.volume(vol)
            if padding > 0:
                voldata = np.pad(voldata, [(v, v) for v in padding_voxels], 'constant', constant_values=0) 
            shift = np.random.normal(scale=std_voxels, size=3)
            shifted_data = scipy.ndimage.interpolation.shift(voldata, shift)
            moving_data[..., vol] = shifted_data

        self.ivm.add(moving_data, grid=output_grid, name=output_name, make_current=True)



from .widgets import OrientDataWidget, ResampleDataWidget
from .processes import ResampleProcess
from .tests import ResampleDataWidgetTest, ResampleProcessTest

QP_MANIFEST = {
    "widgets" : [OrientDataWidget, ResampleDataWidget],
    "processes" : [ResampleProcess],
    "widget-tests" : [ResampleDataWidgetTest,],
    "process-tests" : [ResampleProcessTest,],
}



"""
Quantiphyse - Histogram widget

Copyright (c) 2013-2018 University of Oxford
"""
from .widget import HistogramWidget
from .process import HistogramProcess
from .process_tests import HistogramProcessTest

QP_MANIFEST = {
    "widgets" : [HistogramWidget],
    "processes" : [HistogramProcess],
    "process-tests" : [HistogramProcessTest],
}



"""
Quantiphyse - Base class for widget self-test framework

Copyright (c) 2013-2018 University of Oxford
"""

import sys
import math
import unittest
import traceback

import numpy as np
import scipy

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import DataGrid, ImageVolumeManagement
from quantiphyse.gui import ViewOptions, ImageView
from quantiphyse.utils import QpException, get_plugins

class WidgetTest(unittest.TestCase):
    """
    Base class for a test module for a QP WidgetTest

    Note that it is necessary to physically show the widget during testing. This is ugly
    but enables checking of visible/invisible components.

    The philosophy of widget testing is 'white box', so we feel entitled to use our knowledge
    of the names given to widget components and other internal structure. This means tests
    are likely to go out of date very quickly if not maintained in line with the widgets
    themselves. However it is difficult to test GUI logic otherwise.

    Because of this, widget test classes should be stored in their respective packages and
    exposed to the test framework using the ``widget-tests`` key in ``QP_MANIFEST``. Tests are then
    run using ``quantiphyse --test-all``
    """
    def setUp(self):
        self.qpe, self.error = False, False
        sys.excepthook = self._exc

        self.ivm = ImageVolumeManagement()
        self.opts = ViewOptions(None, self.ivm)
        self.ivl = ImageView(self.ivm, self.opts)

        get_plugins()
        wclass = self.widget_class()
        if wclass is not None:
            self.w = wclass(ivm=self.ivm, ivl=self.ivl, opts=self.opts)
            self.w.init_ui()
            self.w.activate()
            self.w.show()
        else:
            raise unittest.SkipTest("Plugin not found")

        from . import create_test_data
        create_test_data(self)
        
    def tearDown(self):
        if hasattr(self, "w"):
            self.w.hide()
            
    def processEvents(self):
        """
        Process outstanding QT events, i.e. let handlers for widget
        events that we have triggered run

        This must be run every time a test triggers widget events
        in order for the test to detect the effects
        """
        QtCore.QCoreApplication.instance().processEvents()

    def harmless_click(self, btn):
        """ 
        Click a button and check that it produces no error
        """
        if btn.isEnabled():
            btn.clicked.emit()
        self.processEvents()
        self.assertFalse(self.error)

    def _exc(self, exc_type, value, tb):
        """ 
        Exception handler which simply flags whether a user-exception or an error has been caught 
        """
        self.qpe = issubclass(exc_type, QpException)
        self.error = not self.qpe
        if self.error or "--debug" in sys.argv:
            traceback.print_exception(exc_type, value, tb)
        



"""
Quantiphyse - Base class for logging

Copyright (c) 2013-2018 University of Oxford
"""

import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(name)s %(message)s', datefmt='%H:%M:%S')

def set_base_log_level(level):
    """
    Set the base logging level
    """
    logging.getLogger("quantiphyse").setLevel(level)

class LogSource(object):
    """
    Base class for anything which wants to log messages
    """
    def __init__(self):
        logname = "%s.%s" % (self.__module__, self.__class__.__name__)
        if not logname.startswith("quantiphyse."):
            # Plugins do not come under the quantiphyse namespace but we want them
            # to be ancestors of the generic logger
            logname = "quantiphyse." + logname
        self.logger = logging.getLogger(logname)

    def debug(self, *args, **kwargs):
        """
        Log a debug level message
        """
        self.logger.debug(*args, **kwargs)

    def warn(self, *args, **kwargs):
        """
        Log a warning
        """
        self.logger.warn(*args, **kwargs)
    
    def debug_enabled(self):
        return self.logger.getEffectiveLevel() <= logging.DEBUG


"""
Quantiphyse - Basic classes for Extras

Copyright (c) 2013-2018 University of Oxford
"""
import csv
import six

class Extra(object):
    """
    Base class for things which can be stored in the IVM apart from data sets.

    Essentially the only thing an Extra needs to be able to do is be written
    out as a string. We force subclasses to override this. 

    We also provide a metadata dictionary - ideally extras should write their
    metadata in __str__ but in practice this may not be possible when we want
    the output to be compatible with external programs (e.g. writing out a 
    matrix as TSV)
    """
    def __init__(self, name):
        self.name = name
        self.metadata = {}

    def __str__(self):
        raise NotImplementedError("Subclasses of Extra must implement __str__")

class MatrixExtra(Extra):
    """
    Extra which represents a 2D matrix with optional row and column headers
    """
    def __init__(self, name, arr, row_headers=(), col_headers=()):
        """
        :param name: Extra name
        :param arr: List-of-lists or 2D Numpy array containing matrix data
        :param row_headers: Optional sequence of row headers
        :param col_headers: Optional sequence of column headers
        """
        Extra.__init__(self, name)
        if len(arr) == 0:
            raise ValueError("No matrix data given")
        if row_headers and len(row_headers) != len(arr):
            raise ValueError("Incorrect number of row headers given")
        if col_headers and len(col_headers) != len(arr[0]):
            raise ValueError("Incorrect number of column headers given")

        self.arr = arr
        self.row_headers = list(row_headers)
        self.col_headers = list(col_headers)

    def __str__(self):
        """
        Convert matrix to a string in TSV format
        """
        stream = six.StringIO()
        writer = csv.writer(stream, delimiter='\t', lineterminator='\n')
        if self.col_headers:
            if self.row_headers:
                writer.writerow([" "] + self.col_headers)
            else:
                writer.writerow(self.col_headers)

        for row in self.arr:
            if self.row_headers:
                writer.writerow([" "] + list(row))
            else:
                writer.writerow(list(row))

        return stream.getvalue()

class DataFrameExtra(Extra):
    """
    Extra which represents a Pandas data frame
    """
    def __init__(self, name, df):
        """
        :param name: Extra name
        :param arr: List-of-lists or 2D Numpy array containing matrix data
        :param row_headers: Optional sequence of row headers
        :param col_headers: Optional sequence of column headers
        """
        Extra.__init__(self, name)
        self.df = df

    def __str__(self):
        stream = six.StringIO()
        self.df.to_csv(stream, sep='\t')
        return stream.getvalue()



"""
Copyright (c) 2013-2018 University of Oxford
"""
__contrib__ = {
    "Martin Craig" : "Current maintainer",
    "Benjamin Irving" : "Original author",
    "Michael Chappell" : "",
    "Paula Croal" : "",
}

__acknowledge__ = {
    "Julia Schnabel" : "",
    "Sir Mike Brady" : "",
}



"""
Quantiphyse - Analysis process tests

Copyright (c) 2013-2018 University of Oxford
"""
import os
import unittest

import numpy as np
import pandas as pd

from quantiphyse.processes import Process
from quantiphyse.test import ProcessTest

class AnalysisProcessTest(ProcessTest):
    
    def testCalcVolumes(self):
        yaml = """       
  - CalcVolumes:
        roi: mask
        output-name: mask_vol

  - SaveExtras: 
        mask_vol: mask_vol.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("mask_vol" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "mask_vol.tsv")))

    def testSummaryStatsOldName(self):
        yaml = """
  - OverlayStats:
        data: data_3d
        output-name: testdata_stats

  - SaveExtras: 
        testdata_stats: testdata_stats.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_stats" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "testdata_stats.tsv")))

    def testSummaryStats(self):
        yaml = """
  - DataStatistics:
        data: data_3d
        output-name: testdata_stats

  - SaveExtras: 
        testdata_stats: testdata_stats.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_stats" in self.ivm.extras)

        fname = os.path.join(self.output_dir, "case", "testdata_stats.tsv")
        self.assertTrue(os.path.exists(fname))
        df = pd.read_csv(fname, sep='\t')
        data = df.values
        self.assertEquals(data.shape[0], 5)
        self.assertEquals(data.shape[1], 2)
        self.assertAlmostEquals(data[0, 1], np.mean(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[1, 1], np.median(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[2, 1], np.std(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[3, 1], np.min(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[4, 1], np.max(self.data_3d), delta=0.01)

    def testSummaryStatsMultiple(self):
        yaml = """
  - DataStatistics:
        data: [data_3d, data_4d]
        output-name: testdata_stats

  - SaveExtras: 
        testdata_stats: testdata_stats.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_stats" in self.ivm.extras)

        fname = os.path.join(self.output_dir, "case", "testdata_stats.tsv")
        self.assertTrue(os.path.exists(fname))
        df = pd.read_csv(fname, sep='\t')
        data = df.values
        self.assertEquals(data.shape[0], 5)
        self.assertEquals(data.shape[1], 3)
        self.assertAlmostEquals(data[0, 1], np.mean(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[1, 1], np.median(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[2, 1], np.std(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[3, 1], np.min(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[4, 1], np.max(self.data_3d), delta=0.01)
        self.assertAlmostEquals(data[0, 2], np.mean(self.data_4d), delta=0.01)
        self.assertAlmostEquals(data[1, 2], np.median(self.data_4d), delta=0.01)
        self.assertAlmostEquals(data[2, 2], np.std(self.data_4d), delta=0.01)
        self.assertAlmostEquals(data[3, 2], np.min(self.data_4d), delta=0.01)
        self.assertAlmostEquals(data[4, 2], np.max(self.data_4d), delta=0.01)

    def testSummaryStatsRoi(self):
        yaml = """
  - KMeans:
        data: data_3d
        n-clusters: 3
        output-name: clusters
        
  - DataStatistics:
        data: [data_3d, data_4d]
        roi: clusters
        output-name: testdata_stats

  - SaveExtras: 
        testdata_stats: testdata_stats.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_stats" in self.ivm.extras)

        fname = os.path.join(self.output_dir, "case", "testdata_stats.tsv")
        self.assertTrue(os.path.exists(fname))
        df = pd.read_csv(fname, sep='\t')
        data = df.values
        self.assertEquals(data.shape[0], 5)
        self.assertEquals(data.shape[1], 7)
        clusters = self.ivm.data["clusters"].raw()

        for roi_region in range(3):
            data_3d = self.data_3d[clusters == roi_region+1]
            data_4d = self.data_4d[clusters == roi_region+1]
            self.assertAlmostEquals(data[0, 1+roi_region], np.mean(data_3d), delta=0.01)
            self.assertAlmostEquals(data[1, 1+roi_region], np.median(data_3d), delta=0.01)
            self.assertAlmostEquals(data[2, 1+roi_region], np.std(data_3d), delta=0.01)
            self.assertAlmostEquals(data[3, 1+roi_region], np.min(data_3d), delta=0.01)
            self.assertAlmostEquals(data[4, 1+roi_region], np.max(data_3d), delta=0.01)
            self.assertAlmostEquals(data[0, 4+roi_region], np.mean(data_4d), delta=0.01)
            self.assertAlmostEquals(data[1, 4+roi_region], np.median(data_4d), delta=0.01)
            self.assertAlmostEquals(data[2, 4+roi_region], np.std(data_4d), delta=0.01)
            self.assertAlmostEquals(data[3, 4+roi_region], np.min(data_4d), delta=0.01)
            self.assertAlmostEquals(data[4, 4+roi_region], np.max(data_4d), delta=0.01)

    def testRadialProfile(self):
        yaml = """ 
  - RadialProfile:
        data: data_3d
        centre: 2, 2, 2
        output-name: testdata_rp

  - SaveExtras: 
        testdata_rp: testdata_rp.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_rp" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "testdata_rp.tsv")))

    def testHistogram(self):
        yaml = """
  - Histogram:
        data: data_3d
        bins: 10
        output-name: testdata_hist
        roi: mask

  - SaveExtras: 
        testdata_hist: testdata_hist.tsv
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_hist" in self.ivm.extras)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "testdata_hist.tsv")))

    def testExecMultiply(self):
        yaml = """
  - Exec:
      test1: data_3d * 3.14159265  
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("test1" in self.ivm.data)

    def testExecDemean(self):
        yaml = """
  - Exec:
      test1: data_3d - np.mean(data_3d)
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("test1" in self.ivm.data)

    def testExecSubtractVolume(self):
        yaml = """
  - Exec:
      test1: data_3d - data_4d[..., 0]
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("test1" in self.ivm.data)

if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Self-test framework

Copyright (c) 2013-2018 University of Oxford
"""

import math

import numpy as np
import scipy

from quantiphyse.data import DataGrid

from .process_test import ProcessTest
from .widget_test import WidgetTest
from .run import run_tests

__all__ = ["ProcessTest", "WidgetTest", "run_tests", "create_test_data"]

def _test_fn(x, y, z, t=None):
    f = math.exp(-(x**2 + 2*y**2 + 3*z**2))
    if t is not None:
        f *= 1-math.cos(t*2*math.pi)
    return f

def create_test_data(obj, shape=(10, 10, 10), nt=20, motion_scale=0.5):
    """
    Create test data

    Creates the following attributes on obj, each a Numpy array

     - grid
     - data_3d
     - data_4d
     - data_4d_moving
     - mask
    """
    shape = list(shape)
    centre = [float(v)/2 for v in shape]

    obj.grid = DataGrid(shape, np.identity(4))
    obj.data_3d = np.zeros(shape, dtype=np.float32)
    obj.data_4d = np.zeros(shape + [nt,], dtype=np.float32)
    obj.data_4d_moving = np.zeros(shape + [nt,], dtype=np.float32)
    obj.mask = np.zeros(shape, dtype=np.int)

    for x in range(shape[0]):
        for y in range(shape[1]):
            for z in range(shape[2]):
                nx = 2*float(x-centre[0])/shape[0]
                ny = 2*float(y-centre[1])/shape[1]
                nz = 2*float(z-centre[2])/shape[2]
                d = math.sqrt(nx**2 + ny**2 + nz**2)
                obj.data_3d[x, y, z] = _test_fn(nx, ny, nz)
                obj.mask[x, y, z] = int(d < 0.5)
                for t in range(nt):
                    ft = float(t)/nt
                    obj.data_4d[x, y, z, t] = _test_fn(nx, ny, nz, ft)
    
    for t in range(nt):
        tdata = obj.data_4d[:, :, :, t]
        shift = np.random.normal(scale=motion_scale, size=3)
        odata = scipy.ndimage.interpolation.shift(tdata, shift)
        obj.data_4d_moving[:, :, :, t] = odata



import unittest

import numpy as np

from quantiphyse.data import DataGrid
from quantiphyse.test.widget_test import WidgetTest
from quantiphyse.processes import Process
from quantiphyse.test import ProcessTest

from .widgets import ResampleDataWidget

NAME = "resampled"
ORDER = 0

class ResampleDataWidgetTest(WidgetTest):

    def widget_class(self):
        return ResampleDataWidget

    def testNoData(self):
        """ User clicks the run button with no data"""
        self.w.run.btn.clicked.emit()
        self.processEvents()
        self.assertFalse(self.error)
        self.assertTrue(self.qpe)

    def testResampleToSelf3d(self):
        """ 3d data resampled to self"""
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        self.w.order.setCurrentIndex(0)
        self.processEvents()

        self.assertEqual(self.w.data.currentText(), "data_3d")
        self.assertEqual(self.w.grid_data.currentText(), "data_3d")
        self.assertEqual(self.w.output_name.value, "data_3d_res")

        self.w.run.btn.clicked.emit()
        self.processEvents()
        
        self.assertFalse(self.error)
        self.assertTrue("data_3d" in self.ivm.data)
        self.assertTrue("data_3d_res" in self.ivm.data)
        
        # Resampled data should match original data
        self.assertTrue(self.ivm.data["data_3d_res"].grid.matches(self.ivm.data["data_3d"].grid))
        self.assertTrue(np.all(self.ivm.data["data_3d_res"].raw() == self.data_3d))
        
    def testResampleToSelf4d(self):
        """ 4d data resampled to self"""
        self.ivm.add(self.data_4d, grid=self.grid, name="data_4d")
        self.w.order.setCurrentIndex(0)
        self.processEvents()

        self.assertEqual(self.w.data.currentText(), "data_4d")
        self.assertEqual(self.w.grid_data.currentText(), "data_4d")
        self.assertEqual(self.w.output_name.value, "data_4d_res")

        self.w.run.btn.clicked.emit()
        self.processEvents()

        self.assertFalse(self.error)
        self.assertTrue("data_4d" in self.ivm.data)
        self.assertTrue("data_4d_res" in self.ivm.data)
        
        # Resampled data should match original data
        self.assertTrue(self.ivm.data["data_4d_res"].grid.matches(self.ivm.data["data_4d"].grid))
        self.assertTrue(np.all(self.ivm.data["data_4d_res"].raw() == self.data_4d))
        
    def testResampleToHiResNN(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        hires_shape = [dim*2 for dim in self.data_3d.shape]
        hires_grid = DataGrid(hires_shape, np.identity(4) / 2)
        hires_data = np.tile(self.data_3d, (2, 2, 2))
        self.ivm.add(hires_data, grid=hires_grid, name="hires_data")

        self.w.order.setCurrentIndex(0)
        self.w.data.setCurrentIndex(self.w.data.findText("data_3d"))
        self.w.grid_data.setCurrentIndex(self.w.grid_data.findText("hires_data"))
        self.processEvents()

        self.assertEqual(self.w.output_name.value, "data_3d_res")
        
        self.w.run.btn.clicked.emit()
        self.processEvents()

        self.assertFalse(self.error)
        self.assertTrue("data_3d_res" in self.ivm.data)
        
        # Resampled data should match original data but at twice the resolution
        self.assertTrue(self.ivm.data["data_3d_res"].grid.matches(hires_grid))
        data_res = self.ivm.data["data_3d_res"].raw()
        
        for x in range(data_res.shape[0]):
            for y in range(data_res.shape[1]):
                for z in range(data_res.shape[2]):
                    nx, ny, nz = int(float(x)/2+0.5), int(float(y)/2+0.5), int(float(z)/2+0.5)
                    if nx < self.data_3d.shape[0] and ny < self.data_3d.shape[1] and nz < self.data_3d.shape[2]:
                        self.assertEqual(data_res[x, y, z], self.data_3d[nx, ny, nz])
                    else:
                        self.assertEqual(data_res[x, y, z], 0)
                            
    def testResampleToHiResLinear(self):
        self.ivm.add(self.data_3d, grid=self.grid, name="data_3d")
        hires_shape = [dim*2 for dim in self.data_3d.shape]
        hires_grid = DataGrid(hires_shape, np.identity(4) / 2)
        hires_data = np.tile(self.data_3d, (2, 2, 2))
        self.ivm.add(hires_data, grid=hires_grid, name="hires_data")

        self.w.order.setCurrentIndex(1)
        self.w.data.setCurrentIndex(self.w.data.findText("data_3d"))
        self.w.grid_data.setCurrentIndex(self.w.grid_data.findText("hires_data"))
        self.processEvents()

        self.assertEqual(self.w.output_name.value, "data_3d_res")
        
        self.w.run.btn.clicked.emit()
        self.processEvents()

        self.assertFalse(self.error)
        self.assertTrue("data_3d_res" in self.ivm.data)
        
        # Resampled data should match original data but at twice the resolution
        self.assertTrue(self.ivm.data["data_3d_res"].grid.matches(hires_grid))
        data_res = self.ivm.data["data_3d_res"].raw()
        
        X = range(self.data_3d.shape[0])
        Y = range(self.data_3d.shape[1])
        Z = range(self.data_3d.shape[2])
        
        for x in range(data_res.shape[0]):
            for y in range(data_res.shape[1]):
                for z in range(data_res.shape[2]):
                    gx, gy, gz = float(x)/2, float(y)/2, float(z)/2
                    from scipy.interpolate import interpn
                    d = interpn((X, Y, Z), self.data_3d, (gx, gy, gz), method="linear", bounds_error=False, fill_value=0)
                    self.assertAlmostEqual(data_res[x, y, z], d[0])

class ResampleProcessTest(ProcessTest):
    
    def testResample3d(self):
        yaml = """       
  - Resample:
      data: data_3d
      order: 1 # Linear interpolation
      grid: data_3d
      output-name: testdata_resampled
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_resampled" in self.ivm.data)

    def testResample4d(self):
        yaml = """
  - Resample:
      data: data_4d
      order: 3 # Cubic interpolation
      grid: data_4d
      output-name: testdata_resampled
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue("testdata_resampled" in self.ivm.data)
   
if __name__ == '__main__':
    unittest.main()



"""
Quantiphyse - Widget which displays list of all data loaded

Copyright (c) 2013-2018 University of Oxford
"""

from .widget import OverviewWidget

QP_MANIFEST = {
    "widgets" : [OverviewWidget],
}




DEFAULT_VIEW_METADATA_MAIN = {
}

DEFAULT_VIEW_METADATA_OVERLAY = {
}

DEFAULT_VIEW_METADATA_ROI = {
    "visible" : True,
    "roi_only" : False,
    "boundary" : "clamp",
    "alpha" : 255,
    "cmap" : "grey",
    "cmap_range" : None,
    "interp_order" : 0,
    "roi-shade" : True,
    "roi-contour" : False,
    "roi-outline_width" : 3.0,
    "z_value" : 1,
}

class OrthoDataView(QtCore.QObject):
    """
    View of a QpData item on a 2D orthographic view
    """
    BOUNDARY_TRANS = 0
    BOUNDARY_CLAMP = 1
    BOUNDARY_LOWERTRANS = 2
    BOUNDARY_UPPERTRANS = 3

    def __init__(self, qpdata, viewbox, **view_metadata):
        """
        :param qpdata: QpData instance
        :param viewbox: pyqtgraph ViewBox instance 
        :param view_metadata: View parameters
        """
        self._qpdata = qpdata
        self._viewbox = viewbox
        self._plane = None
        self._vol = None
        self._qpdata.metadata["DataView"] = dict(view_metadata)
        self._img = MaskableImage()
        self._img.setZValue(-999)
        self._img.setBoundaryMode(self.BOUNDARY_CLAMP)
        self._visible = False
        self._viewbox.addItem(self._img)
        self._histogram.add_img(self._img)

    @property
    def plane(self):
        return self._plane
    
    @plane.setter
    def plane(self, plane):
        if plane != self._plane:
            self._plane = plane
            self.update()

    @property
    def vol(self):
        return self._vol
    
    @vol.setter
    def vol(self, vol):
        if vol != self._vol and vol < self._qpdata.nvols:
            self._vol = vol
            self.update()

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, visible):
        if visible != self._visible:
            self._visible = visible
            self.update()

    def update(self):
        raise NotImplementedError("OrthoDataView instances must implement update()")

def OrthoMainDataView(OrthoDataView):
    """
    Orthographic view for the main data. 
    
    This is always greyscale and at the bottom of the stack
    """
    def update(self):
        # FIXME interp order
        self._img.setVisible(self._slice is not None and self._visible)
        if self._img.isVisible():
            slicedata, slicemask, scale, offset = self._qpdata.slice_data(self._plane, vol=self._vol, interp_order=1)
            self._img.setTransform(QtGui.QTransform(scale[0, 0], scale[0, 1], scale[1, 0], scale[1, 1], offset[0], offset[1]))
            self._img.setImage(slicedata, autoLevels=False)
            self._img.mask = slicemask



"""
Quantiphyse - Dialog box for editing view options

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division, unicode_literals, print_function, absolute_import

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import NumberVList

class ScaleEditDialog(QtGui.QDialog):
    """
    Dialog used by the view options to allow the user to edit the 
    scale of the 4th volume dimension
    """
    def __init__(self, parent=None, scale=[]):
        QtGui.QDialog.__init__(self, parent)
        
        vbox = QtGui.QVBoxLayout()
        label = QtGui.QLabel('<font size="5">Edit Scale</font>')
        vbox.addWidget(label)

        self.table = NumberVList(scale, expandable=False)
        self.table.sig_changed.connect(self.changed)
        vbox.addWidget(self.table)

        self.buttonBox = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok|QtGui.QDialogButtonBox.Cancel)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        vbox.addWidget(self.buttonBox)

        self.setLayout(vbox)

        shortcut = QtGui.QShortcut(QtGui.QKeySequence.Paste, self.table)
        shortcut.activated.connect(self.paste)

    def paste(self):
        clipboard = QtGui.QApplication.clipboard()
        text = clipboard.text()
        scale = text.strip().split(",")
        if len(scale) != self.table.rowCount():
            scale = text.strip().split()
        if len(scale) != self.table.rowCount():
            scale = text.strip().split("\t")
        if len(scale) == self.table.rowCount():
            try:
                self.table.setValues([float(v) for v in scale])
            except:
                pass

    def changed(self):
        self.buttonBox.button(QtGui.QDialogButtonBox.Ok).setEnabled(self.table.valid())

class ViewOptions(QtGui.QDialog):
    """
    This class is both a dialog to edit viewing options, but also
    the storage for the option values. For now this is convenient,
    however it will probably be necessary to separate the two 
    as the options become more extensive
    """
    RADIOLOGICAL = 0
    NEUROLOGICAL = 1
    
    DATA_ON_TOP = 0
    ROI_ON_TOP = 1

    SHOW = 0
    HIDE = 1

    sig_options_changed = QtCore.Signal(object)

    def __init__(self, parent, ivm):
        super(ViewOptions, self).__init__(parent)
        self.setWindowTitle("View Options")
        #self.setFixedSize(300, 300)

        self.ivm = ivm
        self.ivm.sig_main_data.connect(self.vol_changed)

        # Options
        self.orientation = self.RADIOLOGICAL
        self.crosshairs = self.SHOW
        self.t_type = "Volume"
        self.t_unit = ""
        self.t_scale_type = 0
        self.t_res = 1.0
        self.t_scale = [0, ]
        self.display_order = self.ROI_ON_TOP
        self.interp_order = 0

        grid = QtGui.QGridLayout()
        label = QtGui.QLabel('<font size="5">View Options</font>')
        grid.addWidget(label, 0, 0)

        grid.addWidget(QtGui.QLabel("Orientation"), 2, 0)
        c = QtGui.QComboBox()
        c.addItem("Radiological (Right is Left)")
        c.addItem("Neurological (Left is Left)")
        c.setCurrentIndex(self.orientation)
        c.currentIndexChanged.connect(self.orientation_changed)
        grid.addWidget(c, 2, 1)

        grid.addWidget(QtGui.QLabel("Crosshairs"), 3, 0)
        c = QtGui.QComboBox()
        c.addItem("Show")
        c.addItem("Hide")
        c.setCurrentIndex(self.crosshairs)
        c.currentIndexChanged.connect(self.crosshairs_changed)
        grid.addWidget(c, 3, 1)

        grid.addWidget(QtGui.QLabel("4D Type"), 4, 0)
        self.t_type_edit = QtGui.QLineEdit(self.t_type)
        self.t_type_edit.editingFinished.connect(self.t_type_changed)
        grid.addWidget(self.t_type_edit, 4, 1)
        
        grid.addWidget(QtGui.QLabel("4D Unit"), 5, 0)
        self.t_unit_edit = QtGui.QLineEdit(self.t_unit)
        self.t_unit_edit.editingFinished.connect(self.t_unit_changed)
        grid.addWidget(self.t_unit_edit, 5, 1)
        
        grid.addWidget(QtGui.QLabel("4D Scale"), 6, 0)
        hbox = QtGui.QHBoxLayout()
        self.t_combo = QtGui.QComboBox()
        self.t_combo.addItem("Fixed resolution")
        self.t_combo.addItem("Labelled")
        self.t_combo.setCurrentIndex(self.t_scale_type)
        self.t_combo.currentIndexChanged.connect(self.t_combo_changed)
        hbox.addWidget(self.t_combo)

        self.t_res_edit = QtGui.QLineEdit(str(self.t_res))
        self.t_res_edit.editingFinished.connect(self.t_res_changed)
        hbox.addWidget(self.t_res_edit)

        self.t_btn = QtGui.QPushButton("Edit")
        self.t_btn.setVisible(False)
        self.t_btn.clicked.connect(self.edit_scale)
        hbox.addWidget(self.t_btn)
        grid.addLayout(hbox, 6, 1)

        grid.addWidget(QtGui.QLabel("Display order"), 7, 0)
        c = QtGui.QComboBox()
        c.addItem("Data on top")
        c.addItem("ROI on top")
        c.setCurrentIndex(self.display_order)
        c.currentIndexChanged.connect(self.zorder_changed)
        grid.addWidget(c, 7, 1)

        grid.addWidget(QtGui.QLabel("View interpolation"), 8, 0)
        c = QtGui.QComboBox()
        c.setToolTip("How data is interpolated for display on non-orthogonal grids")
        c.addItem("Nearest neighbour (fast)", 0)
        c.addItem("Linear", 1)
        c.addItem("Cubic spline (slow)", 3)
        c.setCurrentIndex(self.interp_order)
        c.currentIndexChanged.connect(self.interp_changed)
        grid.addWidget(c, 8, 1)

        grid.setRowStretch(9, 1)
        self.setLayout(grid)

    def vol_changed(self, vol):
        """ 
        Do not signal 'options changed', even thought scale points may be updated. 
        The user has not changed any options, and widgets should update themselves 
        to the new volume by connecting to the volume changed signal
        """
        self.update_scale()

    def update_scale(self):
        """
        Update the list of scale points if we have a 4D volume. Always do this if
        we have a uniform scale, if not only do it if the number of points has
        changed (as a starting point for customisation)
        """
        if self.ivm.main is not None and \
           (self.t_scale_type == 0 or self.ivm.main.nvols != len(self.t_scale)):
            self.t_scale = [i*self.t_res for i in range(self.ivm.main.nvols)]

    def orientation_changed(self, idx):
        self.orientation = idx
        self.sig_options_changed.emit(self)

    def crosshairs_changed(self, idx):
        self.crosshairs = idx
        self.sig_options_changed.emit(self)

    def zorder_changed(self, idx):
        self.display_order = idx
        self.sig_options_changed.emit(self)

    def edit_scale(self):
        dlg = ScaleEditDialog(self, self.t_scale)
        if dlg.exec_():
            self.t_scale = dlg.table.values()
        self.sig_options_changed.emit(self)

    def t_unit_changed(self):
        self.t_unit = self.t_unit_edit.text()
        self.sig_options_changed.emit(self)

    def t_type_changed(self):
        self.t_type = self.t_type_edit.text()
        self.sig_options_changed.emit(self)

    def t_res_changed(self):
        self.t_res = float(self.t_res_edit.text())
        self.update_scale()
        self.sig_options_changed.emit(self)
            
    def t_combo_changed(self, idx):
        self.t_scale_type = idx
        self.t_btn.setVisible(idx == 1)
        self.t_res_edit.setVisible(idx == 0)
        self.update_scale()
        self.sig_options_changed.emit(self)

    def interp_changed(self, idx):
        if idx in (0, 1):
            self.interp_order = idx
        else:
            self.interp_order = 3
        self.sig_options_changed.emit(self)



from .widgets import ClusteringWidget, MeanValuesWidget
from .kmeans import KMeansProcess, MeanValuesProcess
from .tests import ClusteringWidgetTest, KMeansProcessTest, MeanValuesProcessTest

QP_MANIFEST = {
    "widgets" : [MeanValuesWidget, ClusteringWidget],
    "widget-tests" : [ClusteringWidgetTest,],
    "process-tests" : [KMeansProcessTest, MeanValuesProcessTest],
    "processes" : [KMeansProcess, MeanValuesProcess],
}



"""
Quantiphyse - ROI builder tools

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import numpy as np
import skimage.segmentation
import scipy.ndimage

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import NumpyData
from quantiphyse.gui.widgets import OverlayCombo, RoiCombo, NumericOption, NumericSlider
from quantiphyse.gui.pickers import PickMode
from quantiphyse.utils import LogSource
from quantiphyse.processes.feat_pca import PcaFeatReduce

class Tool(LogSource):
    """
    An ROI builder tool
    """

    def __init__(self, name, tooltip=""):
        """
        :param name: Name of tool
        :param tooltip: Descriptive text
        """
        LogSource.__init__(self)
        self.name = name
        self.tooltip = tooltip

    def selected(self):
        """
        Called when tool is selected by the user 
        """
        self.ivl.set_picker(PickMode.SINGLE)

    def deselected(self):
        """
        Called when tool is deselected by the user 

        *Must* reliably disconnect any signals connected in the selected() method
        """
        self.ivl.set_picker(PickMode.SINGLE)

    def interface(self):
        """
        Return a user interface 

        :return: QtCore.QLayout suitable for embedding in toolbox
        """
        grid = QtGui.QGridLayout()
        grid.addWidget(QtGui.QLabel(self.tooltip), 0, 0, 1, 2)
        return grid

    def _show_builder_roi(self):
        if self.builder.roiname in self.ivm.rois:
            self.ivm.set_current_roi(self.builder.roiname)

class CrosshairsTool(Tool):
    """
    'Tool' which does nothing but allow the user to navigate the views without
    any side effects
    """
    def __init__(self):
        Tool.__init__(self, "Crosshairs", "Navigate data without adding to ROI")
     
class PickedRegionTool(Tool):
    """
    Base class for tools which identify a picked region which can then be added
    erased or used as a mask
    """

    def __init__(self, name, desc, picker):
        Tool.__init__(self, name, desc)
        self._picker = picker

    def selected(self):
        Tool.selected(self)
        self.ivl.set_picker(self._picker)
   
    def interface(self):
        grid = Tool.interface(self)
        btn = QtGui.QPushButton("Add")
        btn.clicked.connect(self._change(self.builder.ADD))
        grid.addWidget(btn, 1, 0)
        btn = QtGui.QPushButton("Erase")
        btn.clicked.connect(self._change(self.builder.ERASE))
        grid.addWidget(btn, 1, 1)
        btn = QtGui.QPushButton("Mask")
        btn.clicked.connect(self._change(self.builder.MASK))
        grid.addWidget(btn, 2, 0)
        btn = QtGui.QPushButton("Discard")
        btn.clicked.connect(self.selected)
        grid.addWidget(btn, 2, 1)
        return grid

    def _change(self, mode):
        def _cb():
            selected_points = self.ivl.picker.selection(grid=self.builder.grid)
            self.builder.modify(vol=selected_points, mode=mode)
            self.selected()
        return _cb

class PolygonTool(PickedRegionTool):
    """ Tool which selects a polygon region """
    def __init__(self):
        PickedRegionTool.__init__(self, "Polygon", "Select regions using a series of straight lines", PickMode.POLYGON)

class PenTool(PickedRegionTool):
    """ Tool which selects a region drawn around in freehand """
    def __init__(self):
        PickedRegionTool.__init__(self, "Pen", "Draw around ROI region", PickMode.FREEHAND)
     
class RectTool(PickedRegionTool):
    """ Tool which selects a rectangular region """
    def __init__(self):
        PickedRegionTool.__init__(self, "Rectangle", "Click and drag to select rectangular region", PickMode.RECT)

class EllipseTool(PickedRegionTool):
    """ Tool which selects an elliptical region """
    def __init__(self):
        PickedRegionTool.__init__(self, "Ellipse", "Click and drag to select elliptical region", PickMode.ELLIPSE)
       
class PainterTool(Tool):
    """
    Tool which paints voxels when clicked
    """
    def __init__(self, name="Painter", desc="Paint voxels on the ROI"):
        Tool.__init__(self, name, desc)
        self._paint_mode = 1 # FIXME

    def selected(self):
        self.ivl.set_picker(PickMode.PAINT)
        self.ivl.sig_selection_changed.connect(self._point_picked)

    def deselected(self):
        self.ivl.sig_selection_changed.disconnect(self._point_picked)
        Tool.deselected(self)

    def interface(self):
        grid = Tool.interface(self)
        self.size = NumericOption("Brush size", grid, 1, 0, intonly=True, maxval=10, default=1, minval=1, step=2)
        return grid

    def _point_picked(self, picker):
        self.debug("Point picked")
        size = self.size.spin.value()
        points = picker.selection(grid=self.builder.grid)

        grid_axes = self.builder.grid.get_ras_axes()
        gridx = grid_axes[picker.view.xaxis]
        gridy = grid_axes[picker.view.yaxis]
        gridz = grid_axes[picker.view.zaxis]
        self.debug("Grid axes: %i, %i, %i", gridx, gridy, gridz)
        self.debug(str(points))
        grid_points = []
        for point in points:
            for xd in range(-int((size-1)/2), int((size-1)/2)+1, 1):
                for yd in range(-int((size-1)/2), int((size-1)/2)+1, 1):
                    grid_point = [int(v + 0.5) for v in point]
                    grid_point[gridx] = grid_point[gridx] + xd
                    grid_point[gridy] = grid_point[gridy] + yd
                    if not grid_points or grid_point != grid_points[-1]:
                        grid_points.append(grid_point)
        self.debug(str(grid_points))
        self.builder.modify(points=grid_points, mode=self._paint_mode)
        picker.reset()
      
class EraserTool(PainterTool):
    """
    Tool which erases voxels when clicked
    """
    def __init__(self):
        PainterTool.__init__(self, "Eraser", "Remove voxels from the ROI")
        self._paint_mode = 2 # FIXME

class PickTool(Tool):
    """
    Tool which adds entire regions from another ROI
    """
    def __init__(self):
        Tool.__init__(self, "Pick", "Pick regions of an existing ROI")
        self.roi_name = ""
        self.temp_name = "PickToolTempRoi"

    def interface(self):
        grid = Tool.interface(self)

        grid.addWidget(QtGui.QLabel("Existing ROI"), 1, 0)
        self.roi_combo = RoiCombo(self.ivm, none_option=True)
        self.roi_combo.currentIndexChanged.connect(self._existing_roi_changed)
        grid.addWidget(self.roi_combo, 1, 1)

        self.ok_btn = QtGui.QPushButton("Accept")
        self.ok_btn.setEnabled(False)
        self.ok_btn.clicked.connect(self._accepted)
        grid.addWidget(self.ok_btn, 2, 0)
        self.cancel_btn = QtGui.QPushButton("Cancel")
        self.cancel_btn.setEnabled(False)
        self.cancel_btn.clicked.connect(self._reset)
        grid.addWidget(self.cancel_btn, 2, 1)
        self.done_btn = QtGui.QPushButton("Done")
        self.done_btn.setEnabled(False)
        self.done_btn.clicked.connect(self._show_builder_roi)
        grid.addWidget(self.done_btn, 2, 2)
        return grid

    def selected(self):
        self.ivl.set_picker(PickMode.SINGLE)
        self._show_existing_roi()
        self.ivl.sig_selection_changed.connect(self._point_picked)

    def deselected(self):
        self.ivl.sig_selection_changed.disconnect(self._point_picked)
        self._show_builder_roi()
        Tool.deselected(self)

    def _reset(self):
        self.roi_new = None
        self.ivm.delete(self.temp_name)
        self.ok_btn.setEnabled(False)
        self.cancel_btn.setEnabled(False)
        self.done_btn.setEnabled(True)
        self.roi_combo.setEnabled(True)
        self._show_builder_roi()

    def _show_existing_roi(self):
        if self.roi_name in self.ivm.rois:
            self.ivm.set_current_roi(self.roi_name)

    def _existing_roi_changed(self):
        self.roi_name = self.roi_combo.currentText()
        self._show_existing_roi()

    def _accepted(self):
        self.builder.modify(vol=self.roi_new, mode=self.builder.ADD)
        self._reset()

    def _point_picked(self, picker):
        if self.roi_name == "": return

        pos = picker.selection(grid=self.builder.grid)
        roi_picked = self.ivm.rois[self.roi_name]
        picked_region = roi_picked.value(pos, grid=self.builder.grid)

        roi_picked_arr = roi_picked.resample(self.builder.grid).raw()
        self.roi_new = np.zeros(self.builder.grid.shape, dtype=np.int)
        self.roi_new[roi_picked_arr == picked_region] = 1

        self.ivm.add(NumpyData(self.roi_new, grid=self.builder.grid, name=self.temp_name, roi=True), make_current=True)
        self.ok_btn.setEnabled(True)
        self.cancel_btn.setEnabled(True)
        self.done_btn.setEnabled(False)
        self.roi_combo.setEnabled(False)
        
class WalkerTool(Tool):
    """
    Tool which uses the random walker method to select a region

    FIXME this does not work properly in slice mode at the moment because the zaxis/pos 
    given by the picker does not necessarily correspond to the data axes.
    """
    def __init__(self):
        Tool.__init__(self, "Walker", "Automatic segmentation using the random walk algorithm")
        self.segmode = 0
        self.pickmode = 0

    def interface(self):
        grid = Tool.interface(self)

        grid.addWidget(QtGui.QLabel("Source data: "), 1, 0)
        self.ov_combo = OverlayCombo(self.ivm)
        grid.addWidget(self.ov_combo)

        grid.addWidget(QtGui.QLabel("Click to select points: "), 2, 0)
        self.pickmode_combo = QtGui.QComboBox()
        self.pickmode_combo.addItem("Inside the ROI")
        self.pickmode_combo.addItem("Outside the ROI")
        self.pickmode_combo.currentIndexChanged.connect(self._pick_mode_changed)
        grid.addWidget(self.pickmode_combo, 2, 1)
        
        grid.addWidget(QtGui.QLabel("Segmentation mode: "), 3, 0)
        self.segmode_combo = QtGui.QComboBox()
        self.segmode_combo.addItem("Slice")
        self.segmode_combo.addItem("3D")
        self.segmode_combo.currentIndexChanged.connect(self._seg_mode_changed)
        grid.addWidget(self.segmode_combo, 3, 1)

        self.beta = NumericOption("Diffusion difficulty", grid, 4, 0, intonly=True, maxval=20000, default=10000, step=1000)

        btn = QtGui.QPushButton("Segment")
        btn.clicked.connect(self._segment)
        grid.addWidget(btn, 5, 0)
        btn = QtGui.QPushButton("Clear points")
        btn.clicked.connect(self._init)
        grid.addWidget(btn, 5, 1)

        return grid

    def _pick_mode_changed(self, idx=None):
        self.pickmode = idx
        if self.pickmode == 0:
            self.ivl.picker.col = (255, 0, 0)
        else:
            self.ivl.picker.col = (255, 255, 255)

    def _seg_mode_changed(self, idx=None):
        self.segmode = idx
        self._init()

    def _init(self):
        if self.segmode == 1:
            self.ivl.set_picker(PickMode.MULTIPLE)
        else:
            self.ivl.set_picker(PickMode.SLICE_MULTIPLE)
            
        self.labels = np.zeros(self.builder.grid.shape, dtype=np.int)
        self._pick_mode_changed(self.pickmode)

    def selected(self):
        self.pickmode = 0
        self.segmode = 0
        self.ivl.sig_selection_changed.connect(self._points_changed)
        self._init()
        
    def deselected(self):
        self.ivl.sig_selection_changed.disconnect(self._points_changed)
        Tool.deselected(self)

    def _points_changed(self):
        for col, points in self.ivl.picker.selection(grid=self.builder.grid).items():
            if col == ((255, 0, 0)):
                label = 1
            else:
                label = 2

            for pos in points:
                pos = [int(p+0.5) for p in pos]
                # Clamp points to within range of ROI
                for dim in range(3):
                    pos[dim] = min(max(pos[dim], 0), self.labels.shape[dim]-1)
                self.labels[pos[0], pos[1], pos[2]] = label
                
    def _segment(self):
        data = self.ivm.data[self.ov_combo.currentText()].resample(self.builder.grid)
        labels = self.labels

        kwargs = {}
        # Use voxel size correctly
        spacing = [data.grid.spacing[0] / data.grid.spacing[0],
                   data.grid.spacing[0] / data.grid.spacing[1],
                   data.grid.spacing[0] / data.grid.spacing[2]]

        arr = data.raw()
        if arr.ndim > 3:
            # Reduce 4D data to PCA modes
            pca = PcaFeatReduce(n_components=5)
            arr = pca.get_training_features(arr, feature_volume=True)
            kwargs["multichannel"] = True
        else:
            # Normalize data
            arr = (arr / (np.max(arr)-np.min(arr))) + np.min(arr)
            kwargs["multichannel"] = False

        if self.segmode == 0:
            # Segment using 2D slice only
            zaxis = self.ivl.picker.zaxis
            zpos = int(self.ivl.picker.zpos + 0.5)
            #print(zaxis, zpos)
            sl = [slice(None)] * 3
            sl[zaxis] = zpos
            arr = arr[sl]
            labels = self.labels[sl]
            del spacing[zaxis] 
            #print(labels)
            #print(np.count_nonzero(labels))
            #print(np.count_nonzero(arr))

        seg = skimage.segmentation.random_walker(arr, labels, beta=self.beta.spin.value(), 
                                                 mode='cg_mg', spacing=spacing, **kwargs)

        if self.segmode == 0:
            # Create 3D volume using 2D slice
            seg_3d = np.zeros(self.builder.grid.shape, dtype=np.int)
            seg_3d[sl] = seg
            seg = seg_3d

        # Label 2 is used for 'outside region'
        seg[seg == 2] = 0
        
        self.builder.modify(vol=seg, mode=self.builder.ADD)
        self._init()
             
class BucketTool(Tool):
    """
    Tool which performs a thresholded bucket fill
    """

    def __init__(self):
        Tool.__init__(self, "Bucket", "2D or 3D flood fill with thresholding")
        self.point = None
        self.vol = 0

    def interface(self):
        grid = Tool.interface(self)

        #grid.addWidget(QtGui.QLabel("Segmentation mode: "), 3, 0)
        #self.segmode_combo = QtGui.QComboBox()
        #self.segmode_combo.addItem("Slice")
        #self.segmode_combo.addItem("3D")
        #self.segmode_combo.currentIndexChanged.connect(self._dims_changed)
        #grid.addWidget(self.segmode_combo, 3, 1)

        self.uthresh = NumericSlider("Upper threshold", grid, 4, 0, maxval=100, minval=0, default=50, hardmin=True)
        self.uthresh.sig_changed.connect(self._update_roi)
        self.lthresh = NumericSlider("Lower threshold", grid, 5, 0, maxval=0, minval=-100, default=-50, hardmax=True)
        self.lthresh.sig_changed.connect(self._update_roi)
        self.max_tile_size = NumericSlider("Max distance (voxels)", grid, 6, 0, intonly=True, maxval=500, minval=1, default=100, hardmin=True)
        self.max_tile_size.sig_changed.connect(self._update_roi)

        btn = QtGui.QPushButton("Add")
        btn.clicked.connect(self._add)
        grid.addWidget(btn, 7, 0)
        btn = QtGui.QPushButton("Erase")
        btn.clicked.connect(self._erase)
        grid.addWidget(btn, 7, 1)
        btn = QtGui.QPushButton("Mask")
        btn.clicked.connect(self._mask)
        grid.addWidget(btn, 8, 0)
        btn = QtGui.QPushButton("Discard")
        btn.clicked.connect(self._init)
        grid.addWidget(btn, 8, 1)
        return grid

    def _init(self):
        self.ivl.set_picker(PickMode.SINGLE)
        self._point = None
        if "_temp_bucket" in self.ivm.rois:
            self.ivm.delete("_temp_bucket")
        self._show_builder_roi()

    def selected(self):
        self.ivl.sig_selection_changed.connect(self._sel_changed)
        self._init()
        self._sel_changed()
        
    def deselected(self):
        self.ivl.sig_selection_changed.disconnect(self._sel_changed)
        Tool.deselected(self)
        self._init()

    def _sel_changed(self):
        focus = self.ivl.picker.selection(grid=self.builder.grid)
        self.point = [int(v+0.5) for v in focus[:3]]
        self.vol = focus[3]
        self._update_roi()
    
    def _update_roi(self):
        src_data = self.ivm.current_data.resample(self.builder.grid).volume(self.vol)
        focus_value = src_data[self.point[0], self.point[1], self.point[2]]
        thr_hi, thr_lo = focus_value + self.uthresh.value(), focus_value + self.lthresh.value()
        # Heuristic optimization for large data sets. Start by looking at a tile +- 50 voxels
        # around focus. If that contains the region, fine. If not, add 50 voxels to the tile
        # size and go again until we get nothing on the boundary or end up taking the whole
        # data set
        max_tile_size = self.max_tile_size.value()
        tile_size = min(50, max_tile_size)
        while 1:
            tile, offset = self._get_tile(src_data, self.point, tile_size, src_data.shape)
            binarised = ((tile <= thr_hi) & (tile >= thr_lo)).astype(np.int)
            labelled, _ = scipy.ndimage.measurements.label(binarised)
            scipy_label = labelled[self.point[0]-offset[0], self.point[1]-offset[1], self.point[2]-offset[2]]
            labelled[labelled != scipy_label] = 0
            labelled[labelled == scipy_label] = 1
            if labelled.shape == src_data.shape or tile_size == max_tile_size:
                self.debug("Reached full size, breaking: ", labelled.shape, src_data.shape)
                break
            elif (np.count_nonzero(labelled[:, :, 0]) == 0 and 
                  np.count_nonzero(labelled[:, :, -1]) == 0 and
                  np.count_nonzero(labelled[:, 0, :]) == 0 and
                  np.count_nonzero(labelled[:, -1, :]) == 0 and
                  np.count_nonzero(labelled[0, :, :]) == 0 and
                  np.count_nonzero(labelled[-1, :, :]) == 0):
                self.debug("Nothing on boundary, breaking")
                break
            tile_size = min(tile_size + 50, max_tile_size)

        self.roi = np.zeros(self.builder.grid.shape, dtype=np.int)
        tile_shape = labelled.shape
        self.roi[offset[0]:offset[0]+tile_shape[0], offset[1]:offset[1]+tile_shape[1], offset[2]:offset[2]+tile_shape[2]] = labelled
        self.ivm.add(self.roi, name="_temp_bucket", grid=self.builder.grid, roi=True, make_current=True)
        
    def _get_tile(self, arr, centre, size, shape):
        slices, offset = [], []
        for i in range(3):
            start, end = centre[i]-size, centre[i]+size
            if start < 0: start = 0
            if end > shape[i]: end = shape[i]
            slices.append(slice(start, end))
            offset.append(start)
        self.debug("Tile: ", slices, offset)
        return arr[slices], offset

    def _add(self):
        self.builder.modify(vol=self.roi, mode=self.builder.ADD)
        self._init()

    def _erase(self):
        self.builder.modify(vol=self.roi, mode=self.builder.ERASE)
        self._init()

    def _mask(self):
        self.builder.modify(vol=self.roi, mode=self.builder.MASK)
        self._init()

    def _dims_changed(self):
        pass
